@@ -1,5 +1,5 @@
 /*  CorePlayer.cpp - Core player object, most of the hacking is done here!  
- *	Copyright (C) 1998 Andy Lo A Foe <andy@alsa-project.org>
+ *	Copyright (C) 1998-2002 Andy Lo A Foe <andy@alsaplayer.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -14,7 +14,11 @@
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- */ 
+ *
+ *
+ *  $Id: CorePlayer.cpp,v 1.19 2002/04/04 11:48:15 adnans Exp $
+ *
+*/ 
 
 
 /*
@@ -24,11 +28,11 @@
 */
 
 #include "CorePlayer.h"
-#include "Main.h"
 #include <stdio.h>
 #include <string.h>
 #include <assert.h>
 #include <unistd.h>
+#include <dirent.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/time.h>
@@ -37,68 +41,28 @@
 #include <sys/resource.h>
 #include <sys/errno.h>
 #include <fcntl.h>
+#include <dlfcn.h>
 #include <sched.h>
 #include <pthread.h>
 #include <math.h>
-#include <glib.h>
 #include "Effects.h"
+#include "utilities.h"
+#include "alsaplayer_error.h"
 
 #define MAX_FRAGS	16
 #define LOW_FRAGS	4	
 
-// ----------------------------- AMP specific
-
-int CorePlayer::plugin_count;
-input_plugin CorePlayer::plugins[32];
-
-//bool streamer_func(void *arg, void *data, int size);
-
-#ifdef FADER
-static int dest_pitch;
-static int fade_velocity = -1;
-
-void do_fade(int dp, int vp)
-{
-	dest_pitch = dp;
-	fade_velocity = vp;
-}
-
-
-bool fader_func(void *arg, void *data, int size)
-{
-	CorePlayer *obj = (CorePlayer *)arg;
-	if (obj->IsPlaying() && fade_velocity > 0) {
-		int current_pitch = obj->GetSpeed();
-		if (current_pitch != dest_pitch) {
-			if (current_pitch < dest_pitch) { // Pitch up
-				current_pitch = MIN(dest_pitch, current_pitch + 
-					fade_velocity);
-				printf("pitch up %d\n", current_pitch);
-			} else { // Pitch down
-				current_pitch = MAX(dest_pitch, current_pitch -
-					fade_velocity);
-				printf("pitch down %d\n", current_pitch);	
-			}
-			obj->SetSpeed(current_pitch);
-			if (current_pitch == dest_pitch)
-				fade_velocity = -1;
-		}
-	} else {
-		fade_velocity = -1;
-	}	
-	
-	return true;
-}
-#endif
+extern void exit_sighandler(int);
+static char addon_dir[1024];
 
 
 bool surround_func(void *arg, void *data, int size)
 {
 	int amount = 50;
-	gint16 *d = (gint16 *)data;
-	for (gint i=0; i < size; i+=2) {
-		gint16 l =  d[i]; // Left
-		gint16 r =  d[i+1]; // Right
+	int16_t *d = (int16_t *)data;
+	for (int i=0; i < size; i+=2) {
+		int16_t l =  d[i]; // Left
+		int16_t r =  d[i+1]; // Right
 		d[i] =  ((l*(256-amount))-(r*amount))/256;
 		d[i+1] = ((r*(256-amount))-(l*amount))/256;
 	}
@@ -106,7 +70,86 @@ bool surround_func(void *arg, void *data, int size)
 }
 
 
-CorePlayer::CorePlayer(AlsaDAC *the_dac)
+void CorePlayer::Lock()
+{
+	//alsaplayer_error("Getting player lock...");
+	pthread_mutex_lock(&player_mutex);
+	//alsaplayer_error("Got player lock...");
+}
+
+
+void CorePlayer::Unlock()
+{
+	pthread_mutex_unlock(&player_mutex);
+	//alsaplayer_error("Released player lock...");
+}
+
+
+void CorePlayer::load_input_addons()
+{
+	char path[1024];
+	DIR *dir;
+	struct stat buf;
+	dirent *entry;
+	
+	input_plugin_info_type input_plugin_info;
+
+	sprintf(path, "%s/input", addon_dir);
+
+	dir = opendir(path);
+
+	if (dir) {
+		while ((entry = readdir(dir)) != NULL) {
+			if (strcmp(entry->d_name, ".") == 0 ||
+			    strcmp(entry->d_name, "..") == 0) {
+					continue;
+			}
+			sprintf(path, "%s/input/%s", addon_dir, entry->d_name);
+			if (stat(path, &buf)) continue;
+			if (S_ISREG(buf.st_mode)) {
+				void *handle;
+
+				char *ext = strrchr(path, '.');
+
+				if (!ext)
+					continue;
+				
+				ext++;
+
+				if (strcasecmp(ext, "so"))
+					continue;
+
+				if ((handle = dlopen(path, RTLD_NOW |RTLD_GLOBAL))) {
+					input_plugin_info = (input_plugin_info_type)
+						dlsym(handle, "input_plugin_info");
+
+					if (input_plugin_info) {
+#ifdef DEBUG
+						alsaplayer_error("Loading input plugin: %s", path);
+#endif
+						input_plugin *the_plugin = input_plugin_info();
+						if (the_plugin) {
+							the_plugin->handle = handle;
+						}
+						if (!RegisterPlugin(the_plugin)) {
+							alsaplayer_error("Error loading %s", path);
+							dlclose(handle);
+						}
+					} else {
+						alsaplayer_error("Could not find input_plugin_info symbol in shared object `%s'", path);
+						dlclose(handle);
+					}	
+				} else {
+					alsaplayer_error("%s", dlerror());
+				}
+			}	
+		}
+		closedir(dir);
+	}
+}
+
+
+CorePlayer::CorePlayer(AlsaNode *the_node)
 {
 	int i;
 	file_path[0] = 0;
@@ -119,32 +162,24 @@ CorePlayer::CorePlayer(AlsaDAC *the_dac)
 	jumped = jump_point = repitched = write_buf_changed = 0; 
 	new_frame_number = 0;
 	read_direction = DIR_FORWARD;
-	dac = the_dac;
+	node = the_node;
+	pitch = 1.0;
 	sub = NULL;
 	last_read = -1;
 	SetSpeedMulti(1.0);
 	SetSpeed(1.0);
+	SetVolume(100);
+	SetPan(0);
 	buffer = NULL;
 	the_object = NULL;
 
 	if ((buffer = new sample_buf[NR_BUF]) == NULL) {
-#ifdef USE_EXCEPTIONS	
-		throw("Out of memory");
-#else
-		printf("Out of memory in CorePlayer::CorePlayer\n");
+		alsaplayer_error("Out of memory in CorePlayer::CorePlayer");
 		exit(1);
-#endif
 	}
 
 	for (i=0; i < NR_BUF; i++) {
-#ifdef USE_EXCEPTIONS	
-		try {	
-			buffer[i].buf = new SampleBuffer(SAMPLE_STEREO, BUF_SIZE);
-		}
-		catch(...) { throw; }
-#else
 		buffer[i].buf = new SampleBuffer(SAMPLE_STEREO, BUF_SIZE);
-#endif	
 		// Set up next/prev pointers
 		if (i > 0) {
 			buffer[i].prev = &buffer[i-1];
@@ -156,16 +191,23 @@ CorePlayer::CorePlayer(AlsaDAC *the_dac)
 	buffer[0].prev = &buffer[NR_BUF-1];
 	buffer[NR_BUF-1].next = &buffer[0];
 
+	memset(plugins, 0, sizeof(plugins));
+
 	read_buf = write_buf = buffer;
 	pthread_mutex_init(&counter_mutex, NULL); 
 	pthread_mutex_init(&player_mutex, NULL);
 
 	the_object = new input_object;
+#ifndef NEW_PLAY	
 	the_object->write_buffer = NULL;
+#endif
+	the_object->ready = 0;
 
 	pthread_mutex_init(&the_object->object_mutex, NULL);
-	test_sub = new AlsaSubscriber();
-	test_sub->Subscribe(dac);
+	strcpy(addon_dir, ADDON_DIR);
+
+	// Load the input addons
+	load_input_addons();
 }
 
 
@@ -181,30 +223,56 @@ void CorePlayer::ResetBuffer()
 
 CorePlayer::~CorePlayer()
 {
+	input_plugin *tmp;
 	int i;
 
-	if (IsPlaying())
-		Stop();
+	UnregisterPlugins();
 
 	for (i=0; i < NR_BUF; i++) {
 		delete buffer[i].buf;
 	}
-	delete test_sub;
+	pthread_mutex_destroy(&counter_mutex);
+	pthread_mutex_destroy(&player_mutex);
+	pthread_mutex_destroy(&the_object->object_mutex);
 	delete []buffer;
 	delete the_object;
 }
 
 
+void CorePlayer::UnregisterPlugins()
+{
+	input_plugin *tmp;
+	
+	Stop();
+
+	Lock();
+	for (int i = 0; i < plugin_count; i++) {
+		tmp = &plugins[i];
+		//alsaplayer_error("Unloading Input plugin: %s", tmp->name); 
+		if (tmp->handle) {
+			tmp->shutdown(); // Shutdown plugin
+			dlclose(tmp->handle);
+			tmp->handle = NULL;
+			tmp = NULL;
+		}	
+	}		
+	Unlock();
+}
+
 int CorePlayer::RegisterPlugin(input_plugin *the_plugin)
 {
 	int version;
-	// Quick and dirty!
-	input_plugin *tmp = &plugins[plugin_count];
+	input_plugin *tmp;
+	
+	Lock();
+	tmp = &plugins[plugin_count];
 	tmp->version = the_plugin->version;
 	if (tmp->version) {
 		if ((version = tmp->version) != INPUT_PLUGIN_VERSION) {
-			fprintf(stderr, "Wrong version number on plugin v%d, wanted v%d\n",
-				version - 0x1000, INPUT_PLUGIN_VERSION - 0x1000);	
+			alsaplayer_error("Wrong version number on plugin v%d, wanted v%d",
+					version - INPUT_PLUGIN_BASE_VERSION,
+					INPUT_PLUGIN_VERSION - INPUT_PLUGIN_BASE_VERSION);      
+			Unlock();
 			return 0;
 		}
 	}
@@ -222,13 +290,15 @@ int CorePlayer::RegisterPlugin(input_plugin *the_plugin)
 	tmp->sample_rate = the_plugin->sample_rate;
 	tmp->channels = the_plugin->channels;
 	tmp->stream_info = the_plugin->stream_info;
+	tmp->shutdown = the_plugin->shutdown;
 	tmp->nr_tracks = the_plugin->nr_tracks;
 	tmp->track_seek = the_plugin->track_seek;
 	plugin_count++;
 	if (plugin_count == 1) { // First so assign plugin
 		plugin = tmp;
 	}
-	fprintf(stdout, "Input plugin: %s\n", tmp->name);
+	//alsaplayer_error("Loading Input plugin: %s", tmp->name);
+	Unlock();
 	return 1;
 }
 
@@ -237,12 +307,12 @@ unsigned long CorePlayer::GetCurrentTime(int frame)
 {
 	unsigned long result = 0;
 
-	pthread_mutex_lock(&player_mutex);
-	if (plugin && the_object) {
+	Lock();
+	if (plugin && the_object && the_object->ready) {
 		result = plugin->frame_to_sec(the_object, frame < 0 ? GetPosition()
 			: frame);
 	}
-	pthread_mutex_unlock(&player_mutex);
+	Unlock();
 
 	return result;
 }
@@ -252,13 +322,13 @@ int CorePlayer::GetPosition()
 {
 	if (jumped)	// HACK!
 		return jump_point;
-	if (read_buf && plugin && the_object) {
+	if (read_buf && plugin && the_object && the_object->ready) {
 		if (read_buf->start < 0)
 			return 0;		
 
 		int frame_size = plugin->frame_size(the_object);
 		if (frame_size) 
-			return (read_buf->start + (read_buf->buf->GetReadIndex() * 4) / frame_size);
+			return (read_buf->start + (read_buf->buf->read_index * 4) / frame_size);
 		else
 			return 0;
 	} else {
@@ -271,10 +341,28 @@ int CorePlayer::GetFrames()
 {
 	int result = 0;
 
-	pthread_mutex_lock(&player_mutex);
-	if (plugin && the_object)
+	Lock();
+	if (plugin && the_object && the_object->ready)
 		result = plugin->nr_frames(the_object);
-	pthread_mutex_unlock(&player_mutex);
+	Unlock();
+
+	return result;
+}
+
+
+int CorePlayer::GetChannels()
+{
+	return 2;	// Hardcoded for now
+}
+
+int CorePlayer::GetFrameSize()
+{
+	int result = 0;
+
+	Lock();
+	if (plugin && the_object && the_object->ready)
+		result = plugin->frame_size(the_object);
+	Unlock();
 
 	return result;
 }
@@ -284,10 +372,10 @@ int CorePlayer::GetSampleRate()
 {
 	int result = 0;
 
-	pthread_mutex_lock(&player_mutex);
-	if (plugin && the_object)
+	Lock();
+	if (plugin && the_object && the_object->ready)
 		result = plugin->sample_rate(the_object);
-	pthread_mutex_unlock(&player_mutex);
+	Unlock();
 
 	return result;
 }	
@@ -297,11 +385,12 @@ int CorePlayer::GetStreamInfo(stream_info *info)
 {
 	int result = 0;
 
-	pthread_mutex_lock(&player_mutex);
+	Lock();
 	if (plugin && plugin->stream_info && info && the_object) {
+		memset(info, 0, sizeof(stream_info)); // Clear struct
 		result = plugin->stream_info(the_object, info);
 	}
-	pthread_mutex_unlock(&player_mutex);
+	Unlock();
 
 	return result;		
 }
@@ -314,26 +403,35 @@ int CorePlayer::GetStreamInfo(stream_info *info)
 bool CorePlayer::Start(int reset)
 {
 	float result;
+	int output_rate;
+	int tries;
 
-	pthread_mutex_lock(&player_mutex);
+	if (!node) {
+		alsaplayer_error("CorePlayer does not have a node");
+		return false;
+	}	
+
+	Lock();
+	Close();
+	
 	// First check if we have a filename to play
 	if (!Open()) {
-		pthread_mutex_unlock(&player_mutex);
+		Unlock();
 		return false;
 	}
 	producing = true; // So Read32() doesn't generate an error
 	streaming = true;
 
 	ResetBuffer();
-	FrameSeek(0);
 	
 	last_read = -1;
+	output_rate = node->SamplingRate();
 	
 	result = plugin->sample_rate(the_object);
-	if (result == OUTPUT_RATE)
+	if (result == output_rate)
 		SetSpeedMulti(1.0);
 	else
-		SetSpeedMulti((float) result / (float) OUTPUT_RATE);
+		SetSpeedMulti((float) result / (float) output_rate);
 	update_pitch();
 	write_buf_changed = 0;
 	read_buf = write_buf;
@@ -351,21 +449,33 @@ bool CorePlayer::Start(int reset)
 
 	pthread_create(&producer_thread, NULL,
 		(void * (*)(void *))producer_func, this);
-	//pthread_detach(producer_thread);
-	//dosleep(50000);	//  Give the producer a head start
-	
+
+	//alsaplayer_error("Prebuffering...");
+	// Wait for up to 5 seconds
+	tries = 500;
+	while (--tries && !AvailableBuffers() && producing) { 
+		//alsaplayer_error("Waiting for buffers...");
+		dosleep(10000);
+	}
+
 	sub = new AlsaSubscriber();
-	sub->Subscribe(dac);
+	if (!sub) {
+		alsaplayer_error("Subscriber creation failed :-(\n");
+		return false;
+	}	
+	sub->Subscribe(node);
+	//alsaplayer_error("Starting streamer thread...");
 	sub->EnterStream(streamer_func, this);
-	pthread_mutex_unlock(&player_mutex);
+	Unlock();
+
 	return true;
 }  
 
 
 void CorePlayer::Stop(int streamer)
 {
-	pthread_mutex_lock(&player_mutex);
-	if (sub && sub->IsStreaming()) {
+	Lock();
+	if (sub) {
         	delete sub;
         	sub = NULL;
 	}
@@ -373,22 +483,27 @@ void CorePlayer::Stop(int streamer)
 	streaming = false;
 	pthread_mutex_unlock(&counter_mutex); // Unblock if needed
 #ifdef DEBUG	
-	printf("Waiting for producer_thread to finish...\n");
+	alsaplayer_error("Waiting for producer_thread to finish...");
 #endif
+	//pthread_cancel(producer_thread);
 	pthread_join(producer_thread, NULL); 
+	pthread_mutex_destroy(&counter_mutex);
 	pthread_mutex_init(&counter_mutex, NULL);
 	producer_thread = 0;
-#ifdef DEBUG	
-	printf("producer_func is gone now...\n");
-#endif
+	
+	//alsaplayer_error("producer_func is gone now...");
+	
 	ResetBuffer();
 	Close();
-	pthread_mutex_unlock(&player_mutex);
+	Unlock();
 }
 
 
 int CorePlayer::SetSpeed(float val)
 {
+	if (val < 0.0 && !CanSeek())
+		val = 0.0; // Do not allow reverse play if we can't seek
+	//alsaplayer_error("new speed = %.2f", val);	
 	pitch_point = val;
 	repitched = 1;
 
@@ -401,6 +516,15 @@ float CorePlayer::GetSpeed()
 }
 
 
+bool CorePlayer::CanSeek()
+{
+	if (plugin && the_object && the_object->ready) {
+		if (the_object->flags & P_SEEK)
+			return true;
+	}
+	return false;		
+}
+
 int CorePlayer::Seek(int index)
 {
 	jump_point = index;
@@ -411,8 +535,8 @@ int CorePlayer::Seek(int index)
 
 int CorePlayer::FrameSeek(int index)
 {
-	if (plugin && the_object)
-		plugin->frame_seek(the_object, index);
+	if (plugin && the_object && the_object->ready)
+		index = plugin->frame_seek(the_object, index);
 	else 
 		index = -1;
 	return index;
@@ -421,48 +545,71 @@ int CorePlayer::FrameSeek(int index)
 
 void CorePlayer::Close()
 {
-	if (plugin && the_object)
+	if (plugin && the_object && the_object->ready) {
+		the_object->ready = 0;
 		plugin->close(the_object);
+	}	
 }
 
 
-void CorePlayer::SetFile(const char *path)
+// Find best plugin to play a file
+input_plugin *
+CorePlayer::GetPlayer(const char *path)
 {
-	strncpy(file_path, path, 1024);
-}
+	// Check we've got a path
+	if (!strlen(path)) {
+		return NULL;
+	}
 
-bool CorePlayer::Open()
-{
-	bool result = false;
 	float best_support = 0.0;
 	input_plugin *best_plugin = NULL;
-	if (!strlen(file_path)) {
-		printf("Empty path\n");
-		return false;
-	}
+
+	// Go through plugins, asking them
 	for (int i = 0; i < plugin_count; i++) {
 		input_plugin *p = plugins + i;
-		float sl = p->can_handle(file_path);
+		float sl = p->can_handle(path);
 		if (sl > best_support) {
 			best_support = sl;
 			best_plugin = p;
 			if (sl == 1.0) break;
 		}
 	}
-	if (best_plugin == NULL || best_support <= 0.0) {
-		printf("No suitable plugin found for %s\n", file_path);
+
+	// Return best plugin, if there is one
+	if(best_support <= 0.0) return NULL;
+	return best_plugin;
+}
+
+bool CorePlayer::Open(const char *path = NULL)
+{
+	bool result = false;
+	input_plugin *best_plugin;
+	
+	if (path == NULL && !strlen(file_path)) {
+		return false;
+	}
+	if (path) {
+		strncpy(file_path, path, 1023);
+		file_path[1023] = 0;
+	}	
+
+	if ((best_plugin = GetPlayer(file_path)) == NULL) {
+		alsaplayer_error("No suitable plugin found for \"%s\"", file_path);
 		return false;
 	}
 
 	plugin = best_plugin;
-	
+
+	memset(the_object, 0, sizeof(the_object));
 	if (plugin->open(the_object, file_path)) {
 		result = true;
 		frames_in_buffer = read_buf->buf->GetBufferSizeBytes(plugin->frame_size(the_object)) / plugin->frame_size(the_object);
 	} else {
 		result = false;
 		frames_in_buffer = 0;
-	}	
+		plugin->close(the_object);
+	}
+	the_object->ready = 1;
 
 	return result;	
 }
@@ -472,9 +619,8 @@ void print_buf(sample_buf *start)
 {
 	sample_buf *c = start;
 	for (int i = 0; i < NR_BUF; i++, c = c->next) {
-		printf("%d ", c->start);
+		alsaplayer_error("%d ", c->start);
 	}
-	printf("\n");	
 }
 
 int CorePlayer::SetDirection(int direction)
@@ -568,11 +714,11 @@ int CorePlayer::Read32(void *data, int size)
 	}
 	int use_read_direction = read_direction;	
 	int *out_buf = (int *)data;
-	int *in_buf = (int *)read_buf->buf->GetSamples();
-	in_buf += read_buf->buf->GetReadIndex();
+	int *in_buf = (int *)read_buf->buf->buffer_data;
+	in_buf += read_buf->buf->read_index;
 	int buf_index = 0;
 	int tmp_index = 0;
-	int check_index = read_buf->buf->GetReadIndex();
+	int check_index = read_buf->buf->read_index;
 	int base_corr = 0;
 	float use_pitch = pitch;
 	
@@ -596,22 +742,22 @@ int CorePlayer::Read32(void *data, int size)
 	if (use_read_direction == DIR_FORWARD) {
 		while (buf_index < size) {
 			tmp_index = (int) ((float)use_pitch*(float)(buf_index-base_corr));
-			if ((check_index + tmp_index) > (read_buf->buf->GetSamplesInBuffer())-1) {
+			if ((check_index + tmp_index) > (read_buf->buf->write_index)-1) {
 				if (read_buf->next->start < 0 ||
 					read_buf->next == write_buf) {
 					if (read_buf->next->start == -2) {
-						//printf("Next in queue (2)\n");
+						//alsaplayer_error("Next in queue (2)");
 						return -2;
 					}	
 					memset(data, 0, size * 4);
-					if (!IsPlaying()) {
-						//printf("blah 1\n");
+					if (!producing) {
+						//alsaplayer_error("blah 1");
 						return -1;
 					}
 					return size;
 				} else if (read_buf->next->start !=
 					read_buf->start + frames_in_buffer) {
-					printf("------ WTF!!? %d - %d\n",
+					alsaplayer_error("------ WTF!!? %d - %d",
 					read_buf->next->start,
 					read_buf->start);
 				}
@@ -619,7 +765,7 @@ int CorePlayer::Read32(void *data, int size)
 				pthread_mutex_unlock(&counter_mutex);
 				read_buf->buf->SetReadDirection(DIR_FORWARD);
 				read_buf->buf->ResetRead();
-				in_buf = (int *)read_buf->buf->GetSamples();
+				in_buf = (int *)read_buf->buf->buffer_data;
 				base_corr = buf_index;
 				check_index = 0;
 				tmp_index = (int)((float)use_pitch*(float)(buf_index-base_corr)); // Recalc
@@ -634,7 +780,7 @@ int CorePlayer::Read32(void *data, int size)
 					read_buf->prev == write_buf) {
 				//	printf("Back (%d %d) ", AvailableBuffers(), read_buf->prev->start);
 				//	print_buf(read_buf->prev);
-					if (!IsPlaying()) {
+					if (!producing) {
 						//printf("blah 2\n");
 						return -1;
 					}
@@ -645,8 +791,8 @@ int CorePlayer::Read32(void *data, int size)
 				pthread_mutex_unlock(&counter_mutex);
 				read_buf->buf->SetReadDirection(DIR_BACK);
 				read_buf->buf->ResetRead();
-				int buf_size = read_buf->buf->GetSamplesInBuffer();
-				in_buf = (int *)read_buf->buf->GetSamples();
+				int buf_size = read_buf->buf->write_index;
+				in_buf = (int *)read_buf->buf->buffer_data;
 				in_buf += (buf_size + (check_index - tmp_index));
 				base_corr = buf_index;
 				check_index = buf_size-1;
@@ -657,7 +803,7 @@ int CorePlayer::Read32(void *data, int size)
 	}
 	read_buf->buf->Seek(check_index + ((use_read_direction == DIR_FORWARD) ?
 		 tmp_index+1 : -((tmp_index+1) > check_index ? check_index : tmp_index+1)));
-	if (!size) printf("Humm, I copied nothing?\n");
+	if (!size) alsaplayer_error("Humm, I copied nothing?");
 	return size;
 }
 
@@ -666,8 +812,10 @@ extern int thack;
 int CorePlayer::pcm_worker(sample_buf *dest, int start, int len)
 {
 	int frames_read = 0;
-	
-	if (dest && the_object) {
+	int bytes_written = 0;
+	char *sample_buffer;
+
+	if (dest && the_object && the_object->ready) {
 		int count = dest->buf->GetBufferSizeBytes(plugin->frame_size(the_object));
 		dest->buf->Clear();
 		if (last_read == start) {
@@ -677,21 +825,26 @@ int CorePlayer::pcm_worker(sample_buf *dest, int start, int len)
 		if (start < 0) {
 			return -1;
 		}
+#ifndef NEW_PLAY		
 		the_object->write_buffer = dest->buf;
-		while (count > 0) {
-			if (!plugin->play_frame(the_object)) {
+#endif		
+		sample_buffer = (char *)dest->buf->buffer_data;
+		while (count > 0 && producing) {
+			if (!plugin->play_frame(the_object, sample_buffer + bytes_written)) {
 				dest->start = start;
 #ifdef DEBUG
-				printf("frames read = %d\n", frames_read);
+				alsaplayer_error("frames read = %d", frames_read);
 #endif
 				return frames_read;
 			} else {
+				bytes_written += plugin->frame_size(the_object); // OPTIMIZE!
 				frames_read++;
 			}
 			count -= plugin->frame_size(the_object);
 		}
 		dest->start = start;
 		last_read = dest->start + frames_read;
+		dest->buf->SetSamples(bytes_written >> 2);
 		return frames_read;
 	} else {
 		return -1;
@@ -704,9 +857,11 @@ void CorePlayer::producer_func(void *data)
 	CorePlayer *obj = (CorePlayer *)data;
 	int frames_read;
 
+#ifndef NEW_PLAY
 	obj->the_object->write_buffer = NULL;
+#endif	
 #ifdef DEBUG
-	printf("Starting new producer_func...\n");
+	alsaplayer_error("Starting new producer_func...");
 #endif	
 	while (obj->producing) {
 		if (obj->write_buf_changed) {
@@ -730,7 +885,7 @@ void CorePlayer::producer_func(void *data)
 				frames_read = obj->pcm_worker(obj->write_buf, obj->write_buf->start);
 				if (frames_read != obj->frames_in_buffer) {
 					if (obj->write_buf->start >= 0)
-						printf("an ERROR occured or EOF (%d)\n",
+						alsaplayer_error("an ERROR occured or EOF (%d)",
 							obj->write_buf->start);
 					obj->write_buf->prev->start = -2;
 					obj->producing = false;
@@ -742,51 +897,72 @@ void CorePlayer::producer_func(void *data)
 			}
 		} else {
 #if 0
-			printf("producer: going to wait for a free buffer\n");
+			alsaplayer_error("producer: going to wait for a free buffer");
 #endif
 			pthread_mutex_lock(&obj->counter_mutex);
 #if 0
-			printf("producer: unblocked\n");
+			alsaplayer_error("producer: unblocked");
 #endif
 		}	
 	}
-#ifdef DEBUG	
-	printf("Exited producer_func (producing = %d)\n", obj->producing);
-#endif
+	//alsaplayer_error("Exitting producer_func (producing = %d)", obj->producing);
+	pthread_exit(NULL);
 	return;
 }
-
-extern gint global_bal;
-extern gint global_vol_left;
-extern gint global_vol_right;
-extern gint global_vol;
-extern gint global_pitch;
-extern gint global_reverb_on;
-extern gint global_reverb_delay;
-extern gint global_reverb_feedback;
-
+extern int global_bal;
+extern int global_vol_left;
+extern int global_vol_right;
+extern int global_vol;
+extern int global_pitch;
+extern int global_reverb_on;
+extern int global_reverb_delay;
+extern int global_reverb_feedback;
 
 bool CorePlayer::streamer_func(void *arg, void *data, int size)
 {
 	CorePlayer *obj = (CorePlayer *)arg;
 	int count;
+	char input_buffer[16384]; // Should be big enough
 
-	if ((count = obj->Read32(data, size >> 2)) >= 0) {
-		if (global_reverb_feedback) {
-                        echo_effect32(data , size,
-                                global_reverb_delay, global_reverb_feedback);
-		}
-		if (global_vol != 100 || global_bal != 100) {
-                        int left, right;
-                        left = global_vol_left * global_vol;
-                        left /= 100;
-                        right = global_vol_right * global_vol;
-                        right /= 100;
-                        volume_effect32(data, count, left, right);
-                }
+	if ((count = obj->Read32(input_buffer, size / sizeof(short))) >= 0) {
+		int v, p, left, right;
+		p = obj->pan;
+		v = obj->volume;
+		if (v != 100 || p != 0) {
+			if (p == 0) {
+				left = right = 100;
+			} else 	if (p > 0) {
+				right = 100;
+				left = 100 - p;
+			} else {
+				left = 100;
+				right = 100 + p;
+			}
+			if (v != 100) {
+				left *= v;
+				left /= 100;
+				right *= v;
+				right /= 100;
+			}
+			//printf("v = %d, p = %d, left = %d, right = %d\n",
+			//	p, v, left, right);
+			volume_effect32(input_buffer, count, left, right);
+		}	
+		// Now add the data to the current stream
+		int16_t *in_buffer;
+		int16_t *out_buffer;
+		int32_t level;
+		in_buffer = (int16_t *)input_buffer;
+		out_buffer = (int16_t *)data;
+		for (int i=0; i < size ; i++) {
+				level = *(in_buffer++) + *out_buffer;
+				if (level > 32767) level = 32767;
+				else if (level < -32768) level = -32768;
+				*(out_buffer++) = (int16_t) level;
+		}		
 		return true;
 	} else {
-		//printf("Exiting from streamer_func...\n");
+		//alsaplayer_error("Exiting from streamer_func...");
 		obj->streaming = false;
 		return false;
 	}	
